// SPDX-License-Identifier: GPL-2.0
/*
* Copyright (C) 2020 Red Hat. All rights reserved.
*
* This file is released under the GPL.
*/

#include "fsck.h"

static unsigned was_progress = 0;

static void clean_up_progress(void)
{
	if (was_progress) {
		unsigned i;
		for (i = 0; i < was_progress; i++)
			putc(' ', stderr);
		putc('\r', stderr);
		was_progress = 0;
	}
}

void fsck_log(const char *msg, ...)
{
#pragma omp critical(console)
	{
		va_list va;
		va_start(va, msg);
		clean_up_progress();
		fputs("nvfsck: ", stderr);
		vfprintf(stderr, msg, va);
		putc('\n', stderr);
		va_end(va);
	}
}

void fsck_progress(unsigned pass, uint64_t total, uint64_t partial)
{
#pragma omp critical(console)
	{
		char buffer[64];
		unsigned wp, i;
		if (!total)
			goto ret;
		if (partial > total)
			partial = total;
		wp = snprintf(buffer, sizeof buffer, "pass %u:  %"PRIu64"%%  (%"PRIu64" / %"PRIu64")", pass, partial * 100 / total, partial, total);
		fputs(buffer, stderr);
		for (i = wp; i < was_progress; i++)
			putc(' ', stderr);
		putc('\r', stderr);
		was_progress = wp;
ret:;
	}
}

void fsck_internal_error(const char *msg, ...)
{
	va_list va;
	va_start(va, msg);
	fprintf(stderr, "nvfsck internal error: ");
	vfprintf(stderr, msg, va);
	fprintf(stderr, "\n");
	va_end(va);
	abort();
}

static int get_response(void)
{
	int c, d;
	c = getchar();
	d = c;
	while (d != '\n' && d != EOF)
		d = getchar();
	return c;
}

bool fsck_query(struct qkey *q, int level, const char *msg, const char *question, ...)
{
	bool retval;

	if (q->auto_count >= MAX_AUTO_MESSAGES)
		return q->auto_response;

#pragma omp critical(console)
	{
		int c;
		va_list va;

		if (q->auto_count >= MAX_AUTO_MESSAGES) {
			retval = q->auto_response;
			goto ret;
		}

print_again:
		va_start(va, question);
		clean_up_progress();
		fputs("nvfsck: ", stderr);
		vfprintf(stderr, msg, va);
		if (q->auto_count == MAX_AUTO_MESSAGES - 1)
			fputs(" (suppressing similar messages)", stderr);
		putc('\n', stderr);
		va_end(va);

		if (level == Q_JUST_PRINT) {
			q->auto_count++;
			goto ret_true_nomod;
		}

		if (level > Q_PREEN && preen) {
			q->auto_count++;
			goto ret_false;
		}
		if (level == Q_NOAUTO && yes) {
			q->auto_count++;
			goto ret_false;
		}
		if (level != Q_NOAUTO && all) {
			q->auto_count++;
			goto ret_true;
		}

		if (ro) {
			q->auto_count++;
			goto ret_false;
		}
		if (yes) {
			q->auto_count++;
			goto ret_true;
		}

		if (level != Q_NOAUTO) {
			if (q->same) {
				q->auto_count++;
				goto ret_true;
			}
		}

		fprintf(stderr, "%s (y/n%s) ", question, level != Q_NOAUTO ? "/same/all" : "");

		c = get_response();
		if (c == EOF)
			goto ret_false;
		if ((c & 0xdf) == 'Y')
			goto ret_true;
		if ((c & 0xdf) == 'N')
			goto ret_false;
		if ((c & 0xdf) == 'S') {
			if (level != Q_NOAUTO)
				q->same = 1;
			goto ret_true;
		}
		if ((c & 0xdf) == 'A') {
			all = 1;
			goto ret_true;
		}
		goto print_again;

ret_false:
		uncorrected_errors = 1;
		q->auto_response = retval = false;
		goto ret;

ret_true:
		modified = 1;
ret_true_nomod:
		q->auto_response = retval = true;
		goto ret;

ret:;
	}

	return retval;
}
