// SPDX-License-Identifier: GPL-2.0
/*
* Copyright (C) 2020 Red Hat. All rights reserved.
*
* This file is released under the GPL.
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/time.h>

static int dev, log;
static struct stat dev_st, log_st;
static unsigned seed;
static void *map;
static struct timeval tv;

struct file_record {
	uint64_t offset;
	uint64_t length;
	char file[8];
	uint64_t line;
	uint8_t data[];
} __attribute__((packed));

struct local_record {
	uint64_t offset;
	uint64_t length;
	char file[8];
	uint64_t line;
	uint8_t *data;
};

static struct local_record *lr;
static size_t lr_size;

static void *xmalloc(size_t s)
{
	void *p;
	p = malloc(s);
	if (!p) perror("malloc"), exit(1);
	return p;
}

static void *xrealloc(void *p, size_t s)
{
	p = realloc(p, s);
	if (!p) perror("realloc"), exit(1);
	return p;
}

static void load_records(void)
{
	struct file_record *fr;

	lr = xmalloc(0);
	lr_size = 0;

	for (fr = map; (void *)fr < map || (void *)fr < map + log_st.st_size; fr = (void *)&fr->data[fr->length]) {
		lr = xrealloc(lr, (lr_size + 1) * sizeof(struct local_record));
		lr[lr_size].offset = fr->offset;
		lr[lr_size].length = fr->length;
		memcpy(lr[lr_size].file, fr->file, sizeof fr->file);
		lr[lr_size].line = fr->line;
		lr[lr_size].data = fr->data;
#if 0
		if (!fr->length) {
			fprintf(stderr, "==== %lu ====\n", lr_size);
		} else {
			size_t i;
			fprintf(stderr, "%lx - %lx	", fr->offset, fr->length);
			for (i = 0; i < fr->length; i++) fprintf(stderr, " %02x", fr->data[i]);
			fprintf(stderr, "\n");
		}
#endif
		lr_size++;
	}
	if ((void *)fr != map + log_st.st_size) fprintf(stderr, "invalid log file\n"), exit(1);
}

static bool do_records_intersect(struct local_record *lr1, struct local_record *lr2)
{
	if (!lr1->length || !lr2->length)
		return true;
	if (lr1->offset >= lr2->offset + lr2->length ||
	    lr1->offset + lr1->length <= lr2->offset)
		return false;
	return true;
}

__attribute__((unused)) static void mix_records(void)
{
	ssize_t i;
	for (i = 0; i < (ssize_t)lr_size; i++) {
		ssize_t j, k, pos;
		struct local_record ls;
		for (j = i; j >= 1; j--) {
			if (do_records_intersect(&lr[i], &lr[j - 1]))
				break;
		}
		for (k = i; k < (ssize_t)lr_size - 1; k++) {
			if (do_records_intersect(&lr[i], &lr[k + 1]))
				break;
		}
		pos = j + random() % (k - j + 1);
		//fprintf(stderr, "swap (%lu, %lu)\n", i, j + pos);
		//continue;
#if 0
		ls = lr[pos];
		lr[pos] = lr[i];
		lr[i] = ls;
#else
		ls = lr[i];
		if (pos < i) {
			memmove(&lr[pos + 1], &lr[pos], (i - pos) * sizeof(struct local_record));
			lr[pos] = ls;
		} else if (pos > i) {
			memmove(&lr[i], &lr[i + 1], (pos - i) * sizeof(struct local_record));
			lr[pos] = ls;
		}
#endif
	}
}

static void cut_records(void)
{
	size_t new_size = random() % (lr_size + 1);
	fprintf(stderr, "cut: %zu/%zu: %.8s:%"PRIuMAX" - %.8s:%"PRIuMAX"\n", new_size, lr_size,
		new_size ? lr[new_size - 1].file : "",
		new_size ? lr[new_size - 1].line : 0,
		new_size < lr_size ? lr[new_size].file : "",
		new_size < lr_size ? lr[new_size].line : 0);
	lr_size = new_size;
}

static void replay_records(void)
{
	size_t i;
	for (i = 0; i < lr_size; i++) {
		ssize_t r;
		struct local_record *l = &lr[i];
		if (!l->length)
			continue;
		r = pwrite(dev, l->data, l->length, l->offset);
		if (r == -1) perror("pwrite"), exit(1);
		if ((size_t)r != l->length) fprintf(stderr, "short write\n"), exit(1);
	}
}

int main(int argc, char *argv[])
{
	if (argc < 3) fprintf(stderr, "args\n"), exit(1);

	dev = open(argv[1], O_RDWR);
	if (dev == -1) perror(argv[1]), exit(1);
	if (fstat(dev, &dev_st)) perror("fstat"), exit(1);
	if (!(S_ISREG(dev_st.st_mode) || S_ISBLK(dev_st.st_mode))) fprintf(stderr, "'%s' must be a file or a block device\n", argv[1]), exit(1);

	log = open(argv[2], O_RDONLY);
	if (log == -1) perror(argv[2]), exit(1);
	if (fstat(log, &log_st)) perror("fstat"), exit(1);
	if (!S_ISREG(log_st.st_mode)) fprintf(stderr, "'%s' must be a file\n", argv[1]), exit(1);

	if (argc < 4) {
		if (gettimeofday(&tv, NULL)) perror("gettimeofday"), exit(1);
		seed = tv.tv_sec ^ tv.tv_usec;
	} else {
		seed = strtoul(argv[3], NULL, 10);
	}
	srandom(seed);
	fprintf(stderr, "replay seed %u\n", seed);

	map = mmap(NULL, log_st.st_size, PROT_READ, MAP_PRIVATE, log, 0);
	if (map == MAP_FAILED) perror("mmap"), exit(1);

	load_records();
	mix_records();
	cut_records();
	replay_records();

	if (munmap(map, log_st.st_size)) perror("munmap"), exit(1);
	if (close(dev)) perror("close"), exit(1);
	if (close(log)) perror("close"), exit(1);

	return 0;
}
